from typing import Callable, Optional, Union
from bs4 import BeautifulSoup
from hrenpack.filework import TextFile


def xml_save(func: Callable):
    def wrapper(self, *args, **kwargs):
        output = func(self, *args, **kwargs)
        # Avoid calling save if the method is save itself or prettify
        if func.__name__ not in ['save', 'prettify']:
            self.save()
        return output
    return wrapper


class XMLParser(BeautifulSoup):
    def __init__(self, markup: str, features=None, builder=None, parse_only=None, from_encoding=None,
                 exclude_encodings=None, element_classes=None, **kwargs):
        super().__init__(markup, features, builder, parse_only, from_encoding,
                         exclude_encodings, element_classes, **kwargs)

    def del_tag(self, item: str):
        tags = self.find_all(item)
        for tag in tags:
            tag.decompose()
        return tags


class ParserMixin:
    parser: XMLParser

    @xml_save
    def append(self, tag):
        return self.parser.append(tag)

    @xml_save
    def childGenerator(self):
        return self.parser.childGenerator()

    @xml_save
    def clear(self, decompose=False):
        return self.parser.clear(decompose)

    @xml_save
    def copy_self(self):
        return self.parser.copy_self()

    @xml_save
    def decode(self, indent_level=None, eventual_encoding='utf-8', formatter='minimal', iterator=None, **kwargs):
        return self.parser.decode(indent_level, eventual_encoding, formatter, iterator, **kwargs)

    @xml_save
    def decode_contents(self, indent_level=None, eventual_encoding='utf-8', formatter='minimal'):
        return self.parser.decode_contents(indent_level, eventual_encoding, formatter)

    @xml_save
    def decompose(self):
        return self.parser.decompose()

    @xml_save
    def encode(self, encoding='utf-8', indent_level=None, formatter='minimal', errors="xmlcharrefreplace"):
        return self.parser.encode(encoding, indent_level, formatter, errors)

    @xml_save
    def encode_contents(self, indent_level=None, encoding='utf-8', formatter='minimal'):
        return self.parser.encode_contents(indent_level, encoding, formatter)

    @xml_save
    def endData(self, container_class=None):
        return self.parser.endData(container_class)

    @xml_save
    def extend(self, tags):
        return self.parser.extend(tags)

    @xml_save
    def extract(self, _self_index=None):
        return self.parser.extract(_self_index)

    @xml_save
    def fetchAllPrevious(self, *args, **kwargs):
        return self.parser.fetchAllPrevious(*args, **kwargs)

    @xml_save
    def fetchNextSiblings(self, *args, **kwargs):
        return self.parser.fetchNextSiblings(*args, **kwargs)

    @xml_save
    def fetchParents(self, *args, **kwargs):
        return self.parser.fetchParents(*args, **kwargs)

    @xml_save
    def fetchPreviousSiblings(self, *args, **kwargs):
        return self.parser.fetchPreviousSiblings(*args, **kwargs)

    @xml_save
    def find(self, name=None, attrs=None, recursive=True, string=None, **kwargs):
        if attrs is None:
            attrs = {}
        return self.parser.find(name, attrs, recursive, string, **kwargs)

    @xml_save
    def find_all(self, name=None, attrs=None, recursive: bool = True, string=None, limit: Optional[int] = None,
                 _stacklevel: int = 2, **kwargs):
        if attrs is None:
            attrs = {}
        return self.parser.findAll(name, attrs, recursive, string, limit, _stacklevel, **kwargs)

    @xml_save
    def findAllNext(self, *args, **kwargs):
        return self.parser.findAllNext(*args, **kwargs)

    @xml_save
    def findAllPrevious(self, *args, **kwargs):
        return self.parser.findAllPrevious(*args, **kwargs)

    @xml_save
    def findChild(self, *args, **kwargs):
        return self.parser.findChild(*args, **kwargs)

    @xml_save
    def findChildren(self, *args, **kwargs):
        return self.parser.findChildren(*args, **kwargs)

    @xml_save
    def findNext(self, *args, **kwargs):
        return self.parser.findNext(*args, **kwargs)

    @xml_save
    def findNextSibling(self, *args, **kwargs):
        return self.parser.findNextSibling(*args, **kwargs)

    @xml_save
    def findNextSiblings(self, *args, **kwargs):
        return self.parser.findNextSiblings(*args, **kwargs)

    @xml_save
    def findParent(self, *args, **kwargs):
        return self.parser.findParent(*args, **kwargs)

    @xml_save
    def findParents(self, *args, **kwargs):
        return self.parser.findParents(*args, **kwargs)

    @xml_save
    def findPrevious(self, *args, **kwargs):
        return self.parser.findPrevious(*args, **kwargs)

    @xml_save
    def findPreviousSibling(self, *args, **kwargs):
        return self.parser.findPreviousSibling(*args, **kwargs)

    @xml_save
    def findPreviousSiblings(self, *args, **kwargs):
        return self.parser.findPreviousSiblings(*args, **kwargs)

    @xml_save
    def find_all_next(self, name=None, attrs=None, string=None, limit=None, _stacklevel: int = 2, **kwargs):
        if attrs is None:
            attrs = {}
        return self.parser.find_all_next(name, attrs, string, limit, _stacklevel, **kwargs)

    @xml_save
    def find_all_previous(self, name=None, attrs=None, string=None, limit=None, _stacklevel: int = 2, **kwargs):
        if attrs is None:
            attrs = {}
        return self.parser.find_all_previous(name, attrs, string, limit, _stacklevel, **kwargs)

    @xml_save
    def find_next(self, name=None, attrs=None, string=None, **kwargs):
        if attrs is None:
            attrs = {}
        return self.parser.find_next(name, attrs, string, **kwargs)

    @xml_save
    def find_next_sibling(self, name=None, attrs=None, string=None, **kwargs):
        if attrs is None:
            attrs = {}
        return self.parser.find_next_sibling(name, attrs, string, **kwargs)

    @xml_save
    def find_next_siblings(self, name=None, attrs=None, string=None, limit=None, _stacklevel=2, **kwargs):
        if attrs is None:
            attrs = {}
        return self.parser.find_next_siblings(name, attrs, string, limit, _stacklevel, **kwargs)

    @xml_save
    def find_parent(self, name=None, attrs=None, **kwargs):
        if attrs is None:
            attrs = {}
        return self.parser.find_parent(name, attrs, **kwargs)

    @xml_save
    def find_parents(self, name=None, attrs=None, limit=None, _stacklevel=2, **kwargs):
        if attrs is None:
            attrs = {}
        return self.parser.find_parents(name, attrs, limit, _stacklevel, **kwargs)

    @xml_save
    def find_previous(self, name=None, attrs=None, string=None, **kwargs):
        if attrs is None:
            attrs = {}
        return self.parser.find_previous(name, attrs, string, **kwargs)

    @xml_save
    def find_previous_sibling(self, name=None, attrs=None, string=None, **kwargs):
        if attrs is None:
            attrs = {}
        return self.parser.find_previous_sibling(name, attrs, string, **kwargs)

    @xml_save
    def find_previous_siblings(self, name=None, attrs=None, string=None, limit=None, _stacklevel=2, **kwargs):
        if attrs is None:
            attrs = {}
        return self.parser.find_previous_siblings(name, attrs, string, limit, _stacklevel, **kwargs)

    @xml_save
    def format_string(self, s, formatter):
        return self.parser.format_string(s, formatter)

    @xml_save
    def formatter_for_name(self, formatter_name):
        return self.parser.formatter_for_name(formatter_name)

    @xml_save
    def get(self, key, default=None):
        return self.parser.get(key, default)

    @xml_save
    def getText(self, separator="", strip=False, types=()):
        return self.parser.getText(separator, strip, types)

    @xml_save
    def get_attribute_list(self, key, default=None):
        return self.parser.get_attribute_list(key, default)

    @xml_save
    def get_text(self, separator="", strip=False, types=()):
        return self.parser.get_text(separator, strip, types)

    @xml_save
    def handle_data(self, data):
        return self.parser.handle_data(data)

    @xml_save
    def handle_endtag(self, name, nsprefix=None):
        return self.parser.handle_endtag(name, nsprefix)

    @xml_save
    def handle_starttag(self, name, namespace, nsprefix, attrs, sourceline=None, sourcepos=None, namespaces=None):
        return self.parser.handle_starttag(name, namespace, nsprefix, attrs, sourceline, sourcepos, namespaces)

    @xml_save
    def has_attr(self, key):
        return self.parser.has_attr(key)

    @xml_save
    def has_key(self, key):
        return self.parser.has_key(key)

    @xml_save
    def index(self, element):
        return self.parser.index(element)

    @xml_save
    def insert(self, position, new_child):
        return self.parser.insert(position, new_child)

    @xml_save
    def insert_after(self, new_child):
        return self.parser.insert_after(new_child)

    @xml_save
    def insert_before(self, new_child):
        return self.parser.insert_before(new_child)

    @xml_save
    def isSelfClosing(self):
        return self.parser.isSelfClosing()

    @xml_save
    def new_string(self, s, subclass=None):
        return self.parser.new_string(s, subclass)

    @xml_save
    def new_tag(self, name, namespace=None, nsprefix=None, attrs=None, sourceline=None, sourcepos=None, string=None):
        return self.parser.new_tag(name, namespace, nsprefix, attrs, sourceline, sourcepos, string)

    @xml_save
    def nextGenerator(self):
        return self.parser.nextGenerator()

    @xml_save
    def nextSiblingGenerator(self):
        return self.parser.nextSiblingGenerator()

    @xml_save
    def object_was_parsed(self, o, parent=None, most_recent_element=None):
        return self.parser.object_was_parsed(o, parent, most_recent_element)

    @xml_save
    def parentGenerator(self):
        return self.parser.parentGenerator()

    @xml_save
    def popTag(self):
        return self.parser.popTag()

    @xml_save
    def prettify(self, encoding=None, formatter='minimal'):
        return self.parser.prettify(encoding, formatter)

    @xml_save
    def previousGenerator(self):
        return self.parser.previousGenerator()

    @xml_save
    def previousSiblingGenerator(self):
        return self.parser.previousSiblingGenerator()

    @xml_save
    def pushTag(self, tag):
        return self.parser.pushTag(tag)

    @xml_save
    def recursiveChildGenerator(self):
        return self.parser.recursiveChildGenerator()

    @xml_save
    def renderContents(self):
        return self.parser.renderContents()

    @xml_save
    def replaceWith(self, new_element):
        return self.parser.replaceWith(new_element)

    @xml_save
    def replaceWithChildren(self):
        return self.parser.replaceWithChildren()

    @xml_save
    def replace_with(self, new_element):
        return self.parser.replace_with(new_element)

    @xml_save
    def replace_with_children(self):
        return self.parser.replace_with_children()

    @xml_save
    def reset(self):
        return self.parser.reset()

    @xml_save
    def select(self, selector, namespaces=None, limit=0, **kwargs):
        return self.parser.select(selector, namespaces, limit, **kwargs)

    @xml_save
    def select_one(self, selector, namespaces=None, **kwargs):
        return self.parser.select_one(selector, namespaces, **kwargs)

    @xml_save
    def setup(self, parent=None, previous_element=None, next_element=None, previous_sibling=None, next_sibling=None):
        return self.parser.setup(parent, previous_element, next_element, previous_sibling, next_sibling)

    @xml_save
    def smooth(self):
        return self.parser.smooth()

    @xml_save
    def string_container(self, base_class=None):
        return self.parser.string_container(base_class)

    @xml_save
    def unwrap(self):
        return self.parser.unwrap()

    @xml_save
    def wrap(self, wrap_inside):
        return self.parser.wrap(wrap_inside)

    @xml_save
    def del_tag(self, item):
        return self.parser.del_tag(item)


class ExtensibleMarkupLanguageFile(ParserMixin, TextFile):
    _default_parser: str = 'xml'

    def __init__(self, path: str, encoding: Union[str, int] = 'utf-8', **kwargs):
        super().__init__(path, encoding, **kwargs)
        self.parser = XMLParser(self.read(), kwargs.get('parser', self._default_parser))

    def save(self):
        """Сохраняет изменения в файл."""
        self.rewrite(self.prettify())


class HyperTextMarkupLanguageFile(ExtensibleMarkupLanguageFile):
    _default_parser = 'html.parser'


from kivy.uix.scrollview import ScrollView as KivyScrollView


class ScrollView(KivyScrollView):
    def __init__(self, parent=None, child=None, do_scroll_x: bool = False, **kwargs):
        super().__init__(do_scroll_x=do_scroll_x, **kwargs)
        if parent is not None:
            parent.add_widget(self)
        if child is not None:
            self.add_widget(child)



def empty_function(*args, **kwargs):
    pass


def function_if(condition: bool, true, false=empty_function, is_lambda: bool = False):
    if condition:
        return true if is_lambda else true()
    else:
        return false if is_lambda else false()


def lambda_generator(func, *args, **kwargs):
    return lambda: func(*args, **kwargs)


def callable_object(arg):
    return lambda: arg



import importlib
from urllib.parse import urlencode
from django.apps import apps
from django.conf import settings
from django.contrib.auth import logout, login
from django.core.exceptions import ImproperlyConfigured
from django.db.models import IntegerChoices, Model
from django.forms import Form
from django.http import Http404
from django.shortcuts import render, redirect
from typing import Union, Optional, Any
from dataclasses import dataclass
from django.urls import reverse, reverse_lazy

from hrenpack import NullStr


class Category:
    def __init__(self, name: str, slug: str):
        self.name = name
        self.slug = slug

    def __str__(self):
        return f'name={self.name}, slug={self.slug}'


class HrenpackDjangoError(Exception):
    pass


class MenuElement:
    """Элемент меню"""
    def __init__(self, title: str, href: str = '/'):
        self.href = href
        self.title = title

    def __str__(self):
        return self.title


def view_dict(title: str, h1_title: NullStr = None, **kwargs) -> dict:
    # kwargs['media_url'] = '/media/'
    kwargs['title'] = title
    kwargs['h1_title'] = title if h1_title is None else h1_title
    return kwargs


def boolean_choices(arg: IntegerChoices):
    return tuple(map(lambda x: (bool(x[0]), x[1]), arg.choices))


def semicolon_plus(model, del_id: bool = True):
    output = dict()
    for field in model._meta.fields:
        verbose_name = field.verbose_name + ':'
        name = field.name
        output[name] = verbose_name
    if del_id:
        del output['id']
    return output


class BooleanChoices(IntegerChoices):
    @property
    def choices(self):
        return boolean_choices(self)


def get_view_app(view):
    return view.__module__.split('.')[0]


# def get_view_base_template(view):
#     return getattr(view, 'base_template_name', getattr(settings, 'BASE_TEMPLATE', 'empty.html'))


def get_view_base_template(view):
    # Добавьте проверку на None и пустую строку
    base_template = getattr(view, 'base_template_name', None)
    if not base_template:
        base_template = getattr(settings, 'BASE_TEMPLATE', 'empty.html')
    return base_template or 'empty.html'  # Гарантирует возврат непустого значения


def add_url_GET(base_url: str, request=None, **params):
    if request is not None:
        params.update(request.GET)
    return f'{base_url}?{urlencode(params)}'


def url_or_reverse(url: str, lazy_mode: bool = True, **kwargs):
    if '/' in url:
        return url
    func = reverse_lazy if lazy_mode else reverse
    return func(url, kwargs=kwargs)


def get_app_inclusion_namespace(app_name: str):
    urlpatterns = importlib.import_module(settings.ROOT_URLCONF).urlpatterns
    for pattern in urlpatterns:
        if hasattr(pattern, 'app_name') and pattern.app_name == app_name:
            return pattern.namespace


# def get_model(setting_name: str):
#     setting_name = setting_name.upper()
#     setting = getattr(settings, setting_name)
#     try:
#         return apps.get_model(setting, require_ready=False)
#     except ValueError:
#         raise ImproperlyConfigured(f"{setting_name} must be of the form 'app_label.model_name'")
#     except LookupError:
#         raise ImproperlyConfigured(f"{setting_name} refers to model '%s' that has not been installed" % setting)



from typing import Optional, Union, Literal
from kivy.core.window import Window
from hrenpack.framework.kivy.widgets import BoxLayout, Label, Button


class FontPercentLabel(Label):
    def __init__(self, text='', parent=None, font_size_cof: float = 0.02, **kwargs):
        super().__init__(text, parent, **kwargs)
        fs = self.font_size
        self.font_size = max(fs, Window.size[0] * font_size_cof)


class FontPercentButton(Button):
    def __init__(self, text: str = '', parent=None, command=None, font_size_cof: float = 0.02, **kwargs):
        super().__init__(text, parent, command, **kwargs)
        fs = self.font_size
        self.font_size = max(fs, Window.size[0] * font_size_cof)



import platform
from hrenpack import ThemeType

if platform.system() == 'Windows':
    import winreg


    class RegistryError(Exception):
        pass


    def remove_registry_keys(hive: str, *keys):
        for key in keys:
            try:
                hive = eval(f'winreg.{hive}')
                subkey = key
                winreg.DeleteKey(hive, subkey)
                print(f"Удален ключ реестра: {key}")
            except OSError as e:
                raise RegistryError(f"Не удалось удалить ключ реестра {key}: {e}")


    def remove_registry_values(hive, subkey, *value_names):
        try:
            with winreg.OpenKey(hive, subkey, 0, winreg.KEY_SET_VALUE) as reg_key:
                for value_name in value_names:
                    try:
                        winreg.DeleteValue(reg_key, value_name)
                    except FileNotFoundError:
                        raise FileNotFoundError(f"Значение {value_name} не найдено в {subkey}")
                    except Exception as e:
                        raise RegistryError(f"Ошибка при удалении {value_name}: {e}")
        except PermissionError:
            raise OSError("Ошибка: Не хватает прав для удаления значений.")
        except Exception as e:
            raise RegistryError(f"Ошибка: {e}")


    def remove_values_in_keys(hive, **value_names):
        """Указывайте значения через **kwargs: ключ=(значение1, значение2) или
        введите _d_i_c_t_={ключ: (значение1, значение2)}"""
        if '_d_i_c_t_' in value_names:
            value_names = value_names['_d_i_c_t_']
        for key in value_names:
            values = value_names[key]
            remove_registry_values(hive, key, *values)


    def get_windows_theme() -> ThemeType:
        """
        Определяет тему Windows (светлая/темная)
        Возвращает: 'dark' или 'light'
        """
        try:
            # Путь к настройкам персонализации в реестре
            key_path = r"SOFTWARE\Microsoft\Windows\CurrentVersion\Themes\Personalize"
            key_name = "AppsUseLightTheme"  # 0 - темная, 1 - светлая

            # Открываем ключ реестра
            key = winreg.OpenKey(
                winreg.HKEY_CURRENT_USER,
                key_path,
                0,
                winreg.KEY_READ
            )

            # Читаем значение
            value, _ = winreg.QueryValueEx(key, key_name)
            winreg.CloseKey(key)

            return 'light' if value == 1 else 'dark'

        except Exception as e:
            print(f"Ошибка при чтении реестра: {e}")
            return 'light'


    def get_system_theme():
        """
        Определяет системную тему Windows
        Возвращает: 'dark' или 'light'
        """
        try:
            key_path = r"SOFTWARE\Microsoft\Windows\CurrentVersion\Themes\Personalize"
            key_name = "SystemUsesLightTheme"  # Системная тема

            key = winreg.OpenKey(
                winreg.HKEY_CURRENT_USER,
                key_path,
                0,
                winreg.KEY_READ
            )

            value, _ = winreg.QueryValueEx(key, key_name)
            winreg.CloseKey(key)

            return 'light' if value == 1 else 'dark'

        except Exception:
            # Если не нашли системную тему, используем тему приложений
            return get_windows_theme()
else:
    raise ImportError('windows_registry is only supported on Windows')



import argparse, sys, functools
from argparse import Namespace
from hrenpack.functionwork import empty_function
from hrenpack.iterwork import multi_in


def naih(method):
    @functools.wraps(method)
    def wrapper(self, args=None, namespace=None):
        if self._no_args_is_help:
            self.no_args_is_help()
        return getattr(super(type(self), self), method.__name__, empty_function)(args, namespace)
    return wrapper


class ArgumentParser(argparse.ArgumentParser):
    def __init__(self, **kwargs):
        self._no_args_is_help = kwargs.pop('no_args_is_help', False)
        self._version = kwargs.pop('version', None)
        self._add_quiet = kwargs.pop('add_quiet', False)
        super().__init__(**kwargs)
        if self._version is not None:
            self.add_argument('--version', '-v', action='version', version=self._version)
        if self._add_quiet:
            self.add_argument('--quiet', '-q', action='store_true')

    def no_args_is_help(self):
        if len(sys.argv) <= 1:
            self.print_help()
            sys.exit(0)

    @naih
    def parse_args(self, args = None, namespace = None) -> Namespace: pass

    @naih
    def parse_known_args(self, args = None, namespace = None) -> tuple[Namespace, list[str]]: pass

    @naih
    def parse_intermixed_args(self, args = None, namespace = None) -> Namespace: pass

    @naih
    def parse_known_intermixed_args(self, args = None, namespace = None) -> tuple[Namespace, list[str]]: pass

    @property
    def logging(self):
        return not multi_in(sys.argv, '--quiet', '-q')

    def log(self, *values, sep: str = ' ', end: str = '\n'):
        if self.logging:
            print(*values, sep=sep, end=end)



from kivy.uix.image import Image as KivyImage
from kivy.graphics.svg import Svg
from svglib.svglib import svg2rlg
from reportlab.graphics import renderPM
from hrenpack.cmd import get_path_without_extension


class Image(KivyImage):
    def path(self):
        return self.source


class SVGImage(Image):
    def __init__(self, path: str, **kwargs):
        super().__init__(**kwargs)
        drawing = svg2rlg(path)
        png_path = get_path_without_extension(path) + '.png'
        renderPM.drawToFile(drawing, png_path, 'PNG')
        self.source = png_path

if __name__ == '__main__':
    print(SVGImage('../../../icons/update.svg').path())



import psutil, subprocess


def is_process_running(process_name):
    for proc in psutil.process_iter(['name']):
        try:
            # Проверяем имя процесса
            if proc.info['name'] == process_name:
                return True
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass
    return False


def kill_process_if_is_running(process_name, *args, **kwargs):
    if is_process_running(process_name):
        subprocess.run(('taskkill', *args), **kwargs)



import warnings
from typing import Union, Literal

stb = Union[str, bool]
tdl = Union[tuple, list, dict]
_any, _all = any, all


def booltest(bull: stb) -> None:
    if bull.lower() == 'true' or bull.lower() == 'false' or bull is True or bull is False:
        pass
    else:
        raise TypeError("Данная функция может работать только с булевыми значениями")


def str_to_bool(input: stb) -> bool:
    if type(input) is bool:
        return input
    elif input.lower() == 'true' or int(input) == 1:
        return True
    elif input.lower() == 'false' or int(input) == 0:
        return False
    else:
        raise TypeError("Данная функция может работать только с булевыми значениями")


def bool_list_count(input: tdl) -> dict:
    def TF(bull: bool, dict_tf: dict):
        if bull:
            dict_tf[True] += 1
        else:
            dict_tf[False] += 1

    TrueFalse = {True: 0, False: 0}
    if type(input) is dict:
        for key in input:
            value = input[key]
            booltest(value)
            TF(value, TrueFalse)
    else:
        for b in input:
            booltest(b)
            TF(b, TrueFalse)
    return TrueFalse


def str_to_bool_soft(input: stb, return_false: bool = False):
    try:
        return str_to_bool(input)
    except TypeError:
        return False if return_false else input


def for_in(variable, mode: Literal['in', '==', 'is'], *values) -> bool:
    if mode == 'in':
        for value in values:
            if value in variable:
                return True
    elif mode == 'is':
        for value in values:
            if value is variable:
                return True
    else:
        for value in values:
            if value == variable:
                return True
    return False


def equals_all(*args) -> bool:
    args = list(args)
    first = args.pop(0)
    for arg in args:
        if first != arg:
            return False
    return True


def any(*args):
    return _any(args)


def all(*args):
    return _all(args)



from datetime import datetime
from pip_setuptools import setup, clean, find_packages, requirements
from setuptools.command.develop import develop

desc = '\n'.join(reversed(('Универсальная библиотека python для большинства задач', 'A universal python library for most tasks')))


class Develop(develop):
    def run(self):
        # Устанавливаем специальную версию для develop
        self.distribution.metadata.version = self.get_dev_version()
        super().run()

    def get_dev_version(self):
        return str(self.distribution.metadata.version) + f'-build-{datetime.now().strftime("%Y-%m-%d-%H-%M-%S")}'


BASE_REQUIREMENTS = requirements('requirements/requirements.txt')
IMAGE_REQUIREMENTS = requirements('requirements/image_requirements.txt')
FLASK_REQUIREMENTS = requirements('requirements/flask_requirements.txt')
FILETYPE_REQUIREMENTS = requirements('requirements/filetype_requirements.txt')
DEV_REQUIREMENTS = requirements('requirements/dev_requirements.txt')
PYSIDE6_REQUIREMENTS = requirements('requirements/pyside6_requirements.txt')
DJANGO_REQUIREMENTS = requirements('requirements/django_requirements.txt')
PYGAME_REQUIREMENTS = requirements('requirements/pygame_requirements.txt')
KIVY_REQUIREMENTS = requirements('requirements/kivy_requirements.txt')
BASE_DEV_REQUIREMENTS = BASE_REQUIREMENTS + DEV_REQUIREMENTS
EXTRA_REQUIREMENTS = (IMAGE_REQUIREMENTS + FLASK_REQUIREMENTS + FILETYPE_REQUIREMENTS +
                      PYSIDE6_REQUIREMENTS + DJANGO_REQUIREMENTS + PYGAME_REQUIREMENTS + KIVY_REQUIREMENTS)
FULL_DEV_REQUIREMENTS = BASE_DEV_REQUIREMENTS + EXTRA_REQUIREMENTS

REQUIREMENTS = dict(
    base=BASE_REQUIREMENTS,
    image=BASE_REQUIREMENTS + IMAGE_REQUIREMENTS,
    flask=BASE_REQUIREMENTS + FLASK_REQUIREMENTS,
    filetype=BASE_REQUIREMENTS + FILETYPE_REQUIREMENTS,
    pyside6=BASE_REQUIREMENTS + PYSIDE6_REQUIREMENTS,
    django=BASE_REQUIREMENTS + DJANGO_REQUIREMENTS,
    pygame=BASE_REQUIREMENTS + PYGAME_REQUIREMENTS,
    kivy=BASE_REQUIREMENTS + KIVY_REQUIREMENTS,
    dev=BASE_DEV_REQUIREMENTS,
    dev_base=BASE_DEV_REQUIREMENTS,
    dev_image=BASE_DEV_REQUIREMENTS + IMAGE_REQUIREMENTS,
    dev_flask=BASE_DEV_REQUIREMENTS + FLASK_REQUIREMENTS,
    dev_filetype=BASE_DEV_REQUIREMENTS + FILETYPE_REQUIREMENTS,
    dev_pyside6=BASE_DEV_REQUIREMENTS + PYSIDE6_REQUIREMENTS,
    dev_django=BASE_DEV_REQUIREMENTS + DJANGO_REQUIREMENTS,
    dev_pygame=BASE_DEV_REQUIREMENTS + PYGAME_REQUIREMENTS,
    dev_kivy=BASE_DEV_REQUIREMENTS + KIVY_REQUIREMENTS,
    all=BASE_REQUIREMENTS + EXTRA_REQUIREMENTS,
    full=BASE_REQUIREMENTS + EXTRA_REQUIREMENTS,
    dev_all=FULL_DEV_REQUIREMENTS,
    dev_full=FULL_DEV_REQUIREMENTS,
)

clean()
setup(
    name='hrenpack',
    version='3.0.0-beta.1',
    author_email='magilyas.doma.09@list.ru',
    author='Маг Ильяс DOMA (MagIlyasDOMA)',
    description=desc,
    license='MIT',
    url='https://github.com/MagIlyasDOMA/hrenpack',
    packages=find_packages(),
    setup_requires=DEV_REQUIREMENTS,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: Microsoft :: Windows :: Windows 11',
        'Operating System :: OS Independent',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'Framework :: Django',
        'Framework :: Django :: 5.2',
        'Natural Language :: English',
        'Natural Language :: Russian',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Security :: Cryptography',
        'Topic :: Text Processing :: Markup',
        'Topic :: Text Processing :: Markup :: HTML',
    ],
    platforms=[
        'Windows',
        'Windows 10',
        'Windows 11',
        'Windows Server 2019+',
        'Linux'
    ],
    project_urls=dict(
        Source='https://github.com/MagIlyasDOMA/hrenpack',
        Documentation='https://magilyasdoma.github.io/hrenpack/documentation.html',
        Changelog='https://github.com/MagIlyasDOMA/hrenpack/blob/main/CHANGELOG.md'
    ),
    python_requires='>=3.10',
    install_requires=BASE_REQUIREMENTS,
    extras_require=REQUIREMENTS,
    include_package_data=True,
    cmdclass=dict(
        develop=Develop,
    ),
)



from math import factorial
from .classes import range_plus


class ArithmeticProgression:
    def __init__(self, first: float, difference: float):
        self.first = first
        self.difference = difference

    def __getitem__(self, n: int):
        if n <= 0:
            raise IndexError('n must be positive')
        if n == 1:
            return self.first
        return self.first + (self.difference * (n - 1))


class GeometricProgression:
    def __init__(self, first: float, denominator: float):
        if first == 0:
            raise ValueError("First cannot be zero")
        if denominator == 0:
            raise ValueError("Denominator cannot be zero")
        self.first = first
        self.denominator = denominator

    def __getitem__(self, n: int):
        if n <= 0:
            raise IndexError('n must be positive')
        if n == 1:
            return self.first
        return self.first * (self.denominator ** (n - 1))


def subfactorial(n: int) -> int:
    if n < 0:
        raise ValueError("n must be positive")
    elif n == 0:
        return 1
    elif n == 1:
        return 0
    result = 0
    for i in range_plus(n):
        result += (-1 ** n) / factorial(i)
    return factorial(n) * result



from typing import Optional
from kivy.core.window import Window
from kivy.properties import NumericProperty
from hrenpack.listwork import dict_keyf
from hrenpack.framework.kivy.widgets import Label


class DynamicLabel(Label):
    font_size = NumericProperty(20)

    def __init__(self, text='', parent=None, mul: float = 0.03, **kwargs):
        super().__init__(text, parent, **kwargs)
        self.mul = mul
        Window.bind(size=self.update_font_size)
        self.update_font_size(num=14)

    def update_font_size(self, *args, mul: Optional[float] = None, num: Optional[int] = None):
        window_width, window_height = Window.size
        if not num:
            self.font_size = max(16, window_width * (mul if mul else self.mul))
        else:
            self.font_size = '14pt'



import logging
from typing import Optional
from hrenpack.listwork import key_in_dict
from hrenpack.kwargswork import get_kwarg
from hrenpack.framework.kivy.widgets import BoxLayout, Label, input


class TextCheckBox(BoxLayout):
    class CheckBox(input.CheckBox):
        @property
        def pos_x(self):
            return self.pos[0]

        @pos_x.setter
        def pos_x(self, value):
            pos_y = self.pos[1]
            self.pos = (value, pos_y)
            self.parent._set_label_pos_x(value + self.parent._difference)  # Ссылка на разницу родителя

    class Label(Label):
        @property
        def pos_x(self):
            return self.pos[0]

        @pos_x.setter
        def pos_x(self, value):
            pos_y = self.pos[1]
            self.pos = (value, pos_y)

    def __init__(self, text: str, parent=None, command=None, checked: bool = False, difference: float = 5, **kwargs):  # Настраиваемая разница
        if key_in_dict(kwargs, 'vert'):
            vert = kwargs['vert']
            del kwargs['vert']
        else:
            vert = False
        orientation = 'vertical' if vert else 'horizontal'
        checkbox_kwargs = get_kwarg(kwargs, 'checkbox_kwargs', {}, False, True)
        label_kwargs = get_kwarg(kwargs, 'label_kwargs', {}, False, True)
        super().__init__(orientation, parent, padding=difference, **kwargs)
        self._difference_ = difference
        self.checkbox = input.CheckBox(self, command, checked, **checkbox_kwargs)
        self.label = Label(text, self, **label_kwargs)
        self.get_selected = self.is_selected

    @property
    def _difference(self):
        return self._difference_

    @_difference.setter
    def _difference(self, value):
        self._difference_ = value
        self.padding = value

    def is_selected(self):
        return self.checkbox.is_selected()

    def set_selected(self, active: Optional[bool] = None):
        self.checkbox.set_selected(active)

    def _set_label_pos_x(self, pos_x: float):
        self.label.pos_x = pos_x + self._difference  # Используйте разницу, заданную в конструкторе


class DoubleWidget(BoxLayout):
    def __init__(self, child1, child2, parent, **kwargs):
        super().__init__('vertical', parent, **kwargs)
        self.child1 = child1
        self.child2 = child2

        self.add_widget(self.child1)

    def switch(self, *args):
        if self.child1 in self.children:
            self.switch2(*args)
        else:
            self.switch1(*args)

    def switch1(self, *args):
        if self.child2 in self.children:
            self.remove_widget(self.child2)
            self.add_widget(self.child1)
        else:
            logging.debug("Уже активирован первый виджет")

    def switch2(self, *args):
        if self.child1 in self.children:
            self.remove_widget(self.child1)
            self.add_widget(self.child2)
        else:
            logging.debug("Уже активирован второй виджет")



from hrenpack.listwork import split_list as join

translit_db = {
	"latin": {
		"а": "a",
		"б": "b",
		"в": "v",
		"г": "g",
		"д": "d",
		"е": "ye",
		"ё": "yo",
		"ж": "j",
		"з": "z",
		"и": "i",
		"й": "y",
		"к": "k",
		"л": "l",
		"м": "m",
		"н": "n",
		"о": "o",
		"п": "p",
		"р": "r",
		"с": "s",
		"т": "t",
		"у": "u",
		"ф": "f",
		"х": "h",
		"ц": "c",
		"ч": "ch",
		"ш": "sh",
		"щ": "sch",
		"ъ": "",
		"ы": "i",
		"ь": "",
		"э": "e",
		"ю": "yu",
		"я": "ya"
	},
	"cyrillic": {
		"a": "а",
		"b": "б",
		"c": "ц",
		"d": "д",
		"e": "э",
		"f": "ф",
		"g": "г",
		"h": "х",
		"i": "и",
		"j": "ж",
		"k": "к",
		"l": "л",
		"m": "м",
		"n": "н",
		"o": "о",
		"p": "п",
		"q": "ку",
		"r": "р",
		"s": "с",
		"t": "т",
		"u": "у",
		"v": "в",
		"w": "в",
		"x": "кс",
		"y": "й",
		"z": "з"
	}
}


def translit(data: str, lat_to_cyr: bool = False, translit_no_letters: bool = True) -> str:
    data = list(data)
    cyrillic = translit_db['cyrillic']
    latin = translit_db['latin']
    if translit_no_letters:
        latin["ъ"] = "-"
        latin["ь"] = "-"
    dictionary = cyrillic if lat_to_cyr else latin
    del cyrillic, latin
    for i in range(len(data)):
        letter = data[i]
        for key in dictionary:
            if letter == key:
                data[i] = dictionary[key]
            if not lat_to_cyr:
                if letter == "Х":
                    data[i] = 'H'
                elif letter == "х":
                    data[i] = 'h'
    return join(data, '')


def slugify(data: str, translit_no_letters: bool = False) -> str:
    data = data.lower()
    data = translit(data, translit_no_letters=translit_no_letters)
    ups = list('!@#$%^&*()+=/\\+`~"?{}[] ')
    ups.append('"')
    downs = list('.,<>\n\t\r')
    for i in ups:
        data = data.replace(i, '-')
    for i in downs:
        data = data.replace(i, '_')
    while data[-1] in ('-', '_'):
        data = data[:-1]
    return data



from django import forms
from django.contrib.auth import forms as auth_forms, password_validation, get_user_model


class PasswordChangeForm(auth_forms.PasswordChangeForm):
    old_password = forms.CharField(
        label="Старый пароль",
        required=False,
        strip=False,
        widget=forms.PasswordInput(
            attrs={"autocomplete": "current-password", "autofocus": True}
        ),
    )

    password1 = forms.CharField(
        label="Новый пароль",
        required=False,
        strip=False,
        widget=forms.PasswordInput(attrs={"autocomplete": "new-password"}),
        help_text=password_validation.password_validators_help_text_html(),
    )

    password2 = forms.CharField(
        label="Подтвердить пароль",
        required=False,
        widget=forms.PasswordInput(attrs={"autocomplete": "new-password"}),
        strip=False,
        help_text="Для подтверждения введите, пожалуйста, пароль ещё раз.",
    )

    def __bool__(self):
        cd = self.cleaned_data
        return not cd['new_password1'] and cd['new_password2']

    def is_valid(self):
        cd = self.cleaned_data
        if cd['password1'] and cd['password2']:
            pass
        elif cd['password1'] or cd['password2']:
            return False
        return super().is_valid()



from typing import Optional
from dotenv import dotenv_values


def dotenv_to_db(driver: str, env_path: Optional[str] = None) -> dict:
    env = dotenv_values(env_path)
    return dict(
        ENGINE=driver,
        NAME=env['DB_NAME'],
        USER=env['DB_USER'],
        PASSWORD=env['DB_PASSWORD'],
        HOST=env['DB_HOST'],
        PORT=env['DB_PORT']
    )



from typing import Literal, Optional, Union
from kivy.uix.boxlayout import BoxLayout as KivyBoxLayout
from kivy.uix.button import Button as KivyButton
from kivy.uix.floatlayout import FloatLayout as KivyFloatLayout
from kivy.uix.label import Label as KivyLabel
from kivy.uix.image import Image as KivyImage
from hrenpack.decorators import args_kwargs
from hrenpack.framework.kivy.mixins import LayoutMixin, WidgetMixin


class BoxLayout(KivyBoxLayout, LayoutMixin):
    @args_kwargs(args_name='children', copy_kwargs=False)
    def __init__(self, orientation: Literal['vertical', 'horizontal'] = 'vertical', parent=None, *children, padding=0,
                 spacing=0, **kwargs):
        super().__init__(orientation=orientation, padding=padding, spacing=spacing, **kwargs)
        if parent is not None:
            parent.add_widget(self)
        if children:
            self.add(*children)

    def cofficient(self, widget,
                   size_percent: Optional[Union[
                       int, float, tuple[float, float], tuple[int, int], list[int, int], list[float, float]]] = None,
                   pos_percent: Optional[Union[
                       int, float, tuple[float, float], tuple[int, int], list[int, int], list[float, float]]] = None,
                   ):
        def f(p):
            if isinstance(p, (int, float)):
                if p <= 0:
                    raise ValueError("Percent must be greater than zero")
                a, b = self.width * (p / 100), self.height * (p / 100)
            else:
                if p[0] <= 0 or p[1] <= 0:
                    raise ValueError("Percent must be greater than zero")
                a, b = self.width * (p[0] / 100), self.height * (p[1] / 100)
            return a, b

        if size_percent is None and pos_percent is None:
            raise ValueError("Either size_percent or pos_percent must be specified")
        if size_percent is not None:
            widget.size_hint = f(size_percent)
        if pos_percent is not None:
            widget.pos_hint = f(pos_percent)


class FloatLayout(KivyFloatLayout, LayoutMixin):
    @args_kwargs(args_name='children', copy_kwargs=False)
    def __init__(self, parent=None, *children, **kwargs):
        super().__init__(**kwargs)
        if parent is not None:
            parent.add_widget(self)
        if children:
            self.add(*children)


class Label(KivyLabel, WidgetMixin):
    def __init__(self, text='', parent=None, **kwargs):
        super().__init__(text=text, **kwargs)
        if parent is not None:
            parent.add_widget(self)

    def set_text(self, text):
        self.text = text


class Button(KivyButton, WidgetMixin):
    def __init__(self, text: str = '', parent=None, command=None, **kwargs):
        super().__init__(text=text, **kwargs)
        if parent is not None:
            parent.add_widget(self)
        if command:
            self.clicked_connect(command)

    def clicked_connect(self, command, **kwargs):
        self.bind(on_press=command, **kwargs)


class IconButton(KivyBoxLayout, WidgetMixin):
    def __init__(self, icon_path: str, text: str = '', parent=None, command=None, **kwargs):
        super(IconButton, self).__init__(**kwargs)
        self.orientation = 'horizontal'

        # Добавление кнопки с иконкой
        self.button = Button(
            text=text,
            size_hint=(None, None),
            size=(50, 50),
            background_normal=icon_path
        )
        self.add_widget(self.button)


class LoggingLabel(Label):
    state: Literal['success', 'warning', 'error'] = 'success'

    def __init__(self, text: str = '', parent=None, *disabling_widgets, disable_error: bool = False, **kwargs):
        super().__init__(text, parent, color=(0, 1, 0), **kwargs)
        self.disabling_widgets = list(disabling_widgets)
        self.disable_error = disable_error
        self.normal = text

    def set_success(self, text: Optional[str] = None):
        if text is None:
            text = self.normal
        self.color = (0, 1, 0)
        self.set_text(text)
        self.state = 'success'
        if self.disable_error:
            self.enable_widgets()

    def set_warning(self, text: str):
        if self.state != 'error':
            self.state = 'warning'
            self.color = (1, 1, 0)
            self.set_text(text)

    def set_error(self, text: str):
        self.color = (1, 0, 0)
        self.state = 'error'
        self.set_text(text)
        if self.disable_error:
            self.disable_widgets()

    def disable_widgets(self):
        for widget in self.disabling_widgets:
            widget.disable()

    def enable_widgets(self):
        for widget in self.disabling_widgets:
            widget.enable()

    def add_disabled(self, *widgets):
        for widget in widgets:
            self.disabling_widgets.append(widget)


class Image(KivyImage):
    def __init__(self, source: str, **kwargs):
        super().__init__(source=source, **kwargs)



from typing import Optional


def is_int(data) -> bool:
    try:
        int(data)
    except ValueError:
        return False
    else:
        return True


def is_float(data) -> bool:
    try:
        float(data)
    except ValueError:
        return False
    else:
        return True


def is_bool(data) -> bool:
    try:
        bool(data)
    except ValueError:
        return False
    else:
        return True


class TypeEdit:
    @staticmethod
    def string(input) -> str:
        return str(input)

    @staticmethod
    def integer(input) -> int:
        return int(input)

    @staticmethod
    def float(input) -> float:
        return float(input)

    @staticmethod
    def boolean(input):
        if input is True or input.lower() == 'true':
            return True
        elif input is False or input.lower() == 'False':
            return False
        else:
            raise ValueError

    def isString(self, input, isString):
        return self.string(input) if isString else input

    def isInt(self, input, isInt):
        return self.integer(input) if isInt else input

    def isFloat(self, input, isFloat):
        return self.float(input) if isFloat else input

    def isBool(self, input, isBool):
        return self.boolean(input) if isBool else input


def isinstance_multi(obj, *types) -> bool:
    return isinstance(obj, types)


def issubclass_multi(obj, *classes) -> bool:
    return issubclass(obj, classes)


def is_object(arg, return_none: bool = True) -> Optional[bool]:
    if isinstance(arg, type):
        return False
    elif isinstance(arg, object):
        return True
    else:
        if return_none:
            return None
        else:
            return True



import os, warnings
from hrenpack.classes import range_plus
from hrenpack.decorators import deprecated
from hrenpack.functionwork import empty_function
from hrenpack.strwork import randstr
from hrenpack.listwork import split_list
from hrenpack.typings import *


def credits():
    print("Hrenpack")
    print("(c) Mag Ilyas DOMA, 2024-2026.")
    print("Licensed under MIT (https://github.com/MagIlyasDOMA/hrenpack/blob/main/LICENSE)")


__version__ = '3.0.0-beta.1'


def print_message_set(value: bool = False):
    global PRINT_MESSAGE
    PRINT_MESSAGE = value


def notwork():
    print("Данная функция находится в разработке и пока не работает")


def sts(word):
    stars = '*' * len(word)
    return stars


def of_utf8(filename, mode='r'):
    file = open(filename, mode, encoding='utf-8')
    return file


def write_a(path, data):
    file = open(path, 'a', encoding='utf-8')
    file.write(f'{str(data)}\n')
    file.close()


def write(path, text):
    file = open(path, 'w', encoding='utf-8')
    file.write(str(text))
    file.close()


def null():
    warnings.warn('This function will be removed in version 3.0.0, use functionwork.empty_function instead', DeprecationWarning, 2)


def switch(variable, case: dict, default=empty_function):
    for key in case:
        func = case[key]
        if variable == key:
            func()
            break
    else:
        default()


def bincode_generator(length: int, isInt: bool = False):
    bincode = ''
    for i in range(length):
        bincode = bincode + randstr(0, 1)
    return int(bincode) if isInt else bincode


def show_help(path_of_document: str, path: str = ''):
    warnings.warn('This function will be removed in version 3.0.0', DeprecationWarning, 2)
    def return_text(pod):
        document = of_utf8(path_of_document)
        data = document.read()
        document.close()
        return data

    text = return_text(path_of_document)

    if path:
        if not os.path.isfile(path):
            raise FileExistsError(
                f'[WinError 183] Невозможно создать новый файл, так как он уже существует: {path}')
        else:
            file = of_utf8(path, 'w')
            file.write(text)
            file.close()
    else:
        print(text)


def switch_return(variable, case: dict, default=None):
    for key in case:
        value = case[key]
        if variable == value:
            output = value
            break
    else:
        output = default
    return output


def string_error(error: Exception):
    return str(error)


def who_called_me():
    import inspect
    current_frame = inspect.currentframe()
    calling_frame = current_frame.f_back
    return inspect.getfile(calling_frame)


def get_resource(path: str):
    """Вызывает ресурс hrenpack. Работает только, если вызывать внутри пакета hrenpack
    :arg path: Принимаются только пути, относительные \\hrenpack\\resources\\
    """
    warnings.warn('This function will be removed in version 3.0.0', DeprecationWarning, stacklevel=2)
    python_path = who_called_me()
    python_list = python_path.split('\\')
    if 'hrenpack' not in python_list:
        raise NotADirectoryError
    hrenpack_index = python_list.index('hrenpack')
    return '\\'.join(python_list[:hrenpack_index + 1]) + '\\resources\\' + path


def one_return(count: int, value=None):
    if count == 1:
        return value
    else:
        output = list()
        for i in range(count):
            output.append(value)
        return tuple(output)


none_tuple = lambda count: one_return(count)
tuple0 = lambda count: one_return(count, 0)
str_tuple = lambda count: one_return(count, '')


def module_is_installed(module_name: str):
    try:
        __import__(module_name)
    except ImportError:
        return False
    else:
        return True


if module_is_installed('django'):
    from hrenpack.framework.django import apps



from django.conf import settings
from django.utils.decorators import classonlymethod
from django.utils.deprecation import MiddlewareMixin
from django.views.generic.base import ContextMixin, TemplateResponseMixin
from hrenpack import NullStr
from hrenpack.encapsulation import add_attrs_from_dict
from hrenpack.framework.django import view_dict, get_view_app


class DataMixin:
    """Использовать в случае несовместимости с классами View и TemplateView"""
    title: str
    h1_title: NullStr = None

    def get_context_data(self, **kwargs):
        try:
            kwargs = super().get_context_data(**kwargs)
        except AttributeError as error:
            raise error
        kwargs['title'] = self.title
        kwargs['h1_title'] = self.h1_title
        return view_dict(**kwargs)


class NonAbstractMixin:
    @classonlymethod
    def as_view(cls, **initkwargs):
        add_attrs_from_dict(cls, title=initkwargs.get('title', None), h1_title=initkwargs.get('h1_title', None))
        return super().as_view(**initkwargs)


class ModelManagerMixin:
    model_manager = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.model_manager is None:
            self.model_manager = self.model


class SuccessURLMixin:
    success_url: str

    def get_success_url(self):
        return self.success_url


class UserAuthorizeMixin:
    authorize: bool = False

    def get_context_data(self, **kwargs):
        kwargs['authorize'] = self.authorize
        return super().get_context_data(**kwargs)

    @classonlymethod
    def as_view(cls, **initkwargs):
        authorize = initkwargs.get('authorize', False)
        if authorize and not cls.authorize:
            cls.authorize = True
        return super().as_view(**initkwargs)


class TemplateViewMixin(ContextMixin, TemplateResponseMixin):
    pass
